// IMPLEMENTATION MODULE m2

#ifndef MODULE_m2
#define MODULE_m2

/*
	Copyright 2002 by Umberto Salsi <salsi@icosaedro.it>
*/


int m2_min(int a, int b)
{
	if( a <= b )
		return a;
	else
		return b;
}


int m2_max(int a, int b)
{
	if( a >= b )
		return a;
	else
		return b;
}


#define m2_abs(i)  abs(i)


void m2_print(STRING *s)
{
	if( s == NULL )  return;
	fwrite(s->s, s->len, 1, stdout);
}


void m2_error(STRING *s)
{
	if( s == NULL )  return;
	fwrite(s->s, s->len, 1, stderr);
}


STRING* m2_input()
{
	char line[1000];
	int len;

	if( fgets(line, 1000, stdin) == NULL )  return 0;
	len = strlen(line);
	if( len > 0 && line[len-1] == '\n' ) len--;
	return new_STRING((unsigned char *) &line, len);
}


STRING *m2_readchar()
{
	int ch;
	unsigned char s[2];

	ch = getchar();
	if( ch == EOF )
		return NULL;
	else {
		s[0] = ch;
		s[1] = 0;
		return new_STRING(s, 1);
	}
}


int m2_stoi(STRING *s)
{
	char *b;

	if( s == NULL )  return 0;
	b = alloca(s->len + 1);
	memcpy(b, s->s, s->len);
	b[s->len] = 0;
	return atoi(b);
}


double m2_stor(STRING *s)
{
	char *b, *end;

	if( s == NULL )  return 0.0;
	b = alloca(s->len + 1);
	memcpy(b, s->s, s->len);
	b[s->len] = 0;
	return strtod(b, &end);
}


#define m2_itor(i) (double)(i)
#define m2_rtoi(r) (int)(r)


#define m2_inc(ip, d) *(ip)+=d


#define m2_time() time(NULL)


/****************************************************************************

	(Modified) Regular Explessions

****************************************************************************/

typedef enum {
	 circsym, dollarsym, barsym, lroundsym, rroundsym, lsquaresym,
	 rsquaresym, lbracesym, rbracesym, questionsym, asteriscsym,
	 plussym, minussym, periodsym, charsym, endsym
} resymbol;

unsigned char *re_begin; /* primo carattere della re */
unsigned char *re_end;   /* ultimo carattere della re +1 */
unsigned char *s_begin;  /* primo carattere della stringa */
unsigned char *s_end;    /* ultimo carattere della stringa +1 */
int match_array;         /* TRUE=match array, FALSE=no match array */
ARRAY **m_a;             /* array dove mettere le subex che matchano */
unsigned char *re;       /* next char to be examined */
unsigned int cc;         /* current char */
resymbol m2_sym;            /* current symbol type */


void __re_err(char *fmt, ...)
// FIXME: use die() or HALT().
{
	va_list ap;
	char *s;

	fflush(stdout);
	fprintf(stderr, "\n**** M2 RUNTIME ERROR:\n");
	fprintf(stderr, "     Regular expression '"); \
	fwrite(re_begin, re_end - re_begin, 1, stderr); \
	fprintf(stderr, "',\n     at char no. %d: ", re - re_begin); \
	va_start(ap, fmt);
	vasprintf(&s, fmt, ap);
	va_end(ap);
	fprintf(stderr, "%s\n", s);
	exit(1);
}


void next()
/*
	Ritorna in m2_sym il prossimo simbolo della *re.
	In cc mette il carattere.
	Se e' un carattere preceduto da \ ritorna m2_sym==charsym e cc contiene
	il carattere.
	Se abbiamo raggiunto la fine della re ritorna m2_sym==endsym e cc==0.
*/
{
	if( re >= re_end ){
		m2_sym = endsym;
		cc = 0;
		re++; /* per il rewind, anche se vado "fuori" */
		return;
	}
	if( *re == '\\' ){
		re++;
		if( re >= re_end )
			__re_err("expected char after `\'");
		m2_sym = charsym;
		cc = *re;
		re++;
		return;
	}
	cc = *re;
	re++;
	switch( cc ){
		case '^': m2_sym = circsym; break;
		case '$': m2_sym = dollarsym; break;
		case '|': m2_sym = barsym; break;
		case '(': m2_sym = lroundsym; break;
		case ')': m2_sym = rroundsym; break;
		case '[': m2_sym = lsquaresym; break;
		case ']': m2_sym = rsquaresym; break;
		case '{': m2_sym = lbracesym; break;
		case '}': m2_sym = rbracesym; break;
		case '?': m2_sym = questionsym; break;
		case '*': m2_sym = asteriscsym; break;
		case '+': m2_sym = plussym; break;
		case '-': m2_sym = minussym; break;
		case '.': m2_sym = periodsym; break;
		default: m2_sym = charsym; break;
	}
}


void parser_init(unsigned char *re_start)
/*
	Inizializza il parser next() per partire dal carattere *re_start
	della re.
*/
{
	re = re_start;
	next();
}


void match_repeater(int *a, int *b)
/*
	Controlla se c'e' un ripetitore a partire dalla posizione corrente
	nella re.
	Mette in *a e *b il range trovato o il default {1,1}.
*/
{
	switch( m2_sym ){

	case questionsym:
		*a = 0; *b = 1; next(); return;
	
	case plussym:
		*a = 1; *b = 9999; next(); return;
	
	case asteriscsym:
		*a = 0; *b = 9999; next(); return;
	
	case lbracesym:
		next();
		*a = 0;
		if( isdigit(cc) ){
			do {
				*a = 10*(*a) + cc - '0';
				if( *a > 9999 )
					__re_err("expected number <= 9999");
				next();
			} while( isdigit(cc) );
		}
		if( cc == ',' ){
			next();
			if( isdigit(cc) ){
				*b = 0;
				do {
					*b = 10*(*b) + cc - '0';
					if( *b > 9999 ) __re_err("expected number <= 9999");
					next();
				} while( isdigit(cc) );
			} else {
				*b = 9999;
			}
		} else {
			*b = *a;
		}
		if( cc != '}' ) __re_err("expected `}'");
		next();
		if( *a > *b ) __re_err("unordered range {%d,%d}", *a, *b);
		return;
	
	default:
		*a = 1; *b = 1; return;
	}
}


int match_continue(unsigned char *s, int *ks, int toplevel);

int match_set(unsigned char *s, int *ks)
/*
	m2_sym == lsquaresym
	Matching di un set ed eventuale repeater.
	Ritorna in *ks il numero di caratteri di s consumati.
	Ritorna 1 se ha successo, 0 se non ha successo.
*/
{
	int i;
	int x;
	int a, b;
	unsigned char set[256];

	//printf("match_set(%s, %s)\n", s, re);

	memset(set, 0, 256);

	next();
	if( m2_sym == circsym ){
		x = 0;
		next();
	} else {
		x = 1;
	}
	do {
		if( m2_sym != charsym ) __re_err("expected literal char in set");
		a = cc;
		set[a] = 1;
		next();
		if( m2_sym == minussym ){ /* range: */
			next();
			if( m2_sym != charsym )
				__re_err("expected upper limit in range");
			b = cc;
			if( a > b )
				__re_err("reversed range [%c-%c]", a, b);
			for( i=a; i<=b; i++ )  set[i] = 1;
			next();
		}
	} while( m2_sym != rsquaresym );
	next();
	
	match_repeater(&a, &b);

	/* conta quanti caratteri da *s in poi sono nel set: */
	i = 0;
	*ks = 0;
	while( s < s_end && set[(int)*s] == x && i < b ){
		i++;
		s++; (*ks)++;
	}

	/* verifica che il conteggio corrisponda a quanto richiesto: */
	if( i >= a )
		return TRUE;
	else
		return FALSE;
}


void match_skip_subex(int *a, int *b)
/*
	m2_sym == lroundsym
	Cerca la fine della subex e ritorna in {*a,*b} il ripetitore.
*/
{
	int level;

	level = 0;
	do {
		if( m2_sym == endsym )
			__re_err("missing teminating subexpression");
		if( m2_sym == lroundsym )
			level++;
		else if ( m2_sym == rroundsym ){
			level--;
		}
		next();
	} while( level > 0 );
	
	match_repeater(a, b);
}


void match_skip_set()
/*
	m2_sym == lsquaresym
	Skip di un set ed eventuale repeater.
*/
{
	int a, b;

	do {
		next();
	} while( m2_sym != rsquaresym && m2_sym != endsym );
	if( m2_sym == endsym )
		__re_err("missing set ending `]'");
	match_repeater(&a, &b);
}
	


void match_skip_branch()
/*
	Esegue lo skip di un branch ed eventuale repeater.
*/
{
	int a, b;

	switch( m2_sym ){

	case lroundsym:
		match_skip_subex(&a, &b);
		return;
	
	case lsquaresym:
		match_skip_set();
		return;
	
	case periodsym:
		next();
		return;
	
	case dollarsym:
		if( re < re_end )
			__re_err("`$' allowed at the end of the RE only");
		return;
	
	case charsym:
		next();
		match_repeater(&a, &b);
		return;
	
	default:
		__re_err("illegal char %c at branch start", cc); /* cc? */
		
	}
}


int match_branch(unsigned char *s, int *ks, int toplevel)
/*
	Matching di un branch ed eventuale repeater.
	Per il parametro 'toplevel' v. la funzione match_continue().
	Ritorna 1 se ha successo, 0 se non ha successo.
*/
{
	int i, c, a, b, ks2;
	unsigned char *sstart;
	unsigned char *substart; /* il char dopo la '(' della subex */
	unsigned char *subend; /* il char dopo la subex ed event. repeater */

	*ks = 0;

	switch( m2_sym ){
	
	case lsquaresym:
		if( ! match_set(s, &ks2) )  return FALSE;
		s += ks2; *ks += ks2;
		return TRUE;

	case lroundsym:
		sstart = s;
		substart = re; /* point at the char after '(' */
		match_skip_subex(&a, &b);
		subend = re; /* point to the char after the branch */
		i = 0;
		while( i < b ){
			/* rewind all'inizio della subex: */
			parser_init(substart);
			if( ! match_continue(s, &ks2, 0) ) break;
			i++;
			s += ks2; *ks += ks2;
		}
		/* imposta *re correttamente prima di tornare: */
		if( *(subend-2) == '\\' ) /* era un char con \ */
			parser_init(subend-2);
		else
			parser_init(subend-1);
		if( i < a )
			return FALSE;
		else {
			if( toplevel && match_array ){
				assign_ARRAY_OF_STRING(m_a, *m_a==NULL? 0:(*m_a)->n_elems,
					new_STRING(sstart, s-sstart));
			}
			return TRUE;
		}

	case dollarsym:
		next();
		if( m2_sym != endsym )
			__re_err("`$' allowed at the end of the RE only");
		if( s >= s_end )
			return TRUE;
		else
			return FALSE;
	
	case periodsym:
		next();
		match_repeater(&a, &b);
		i = 0;
		while( s < s_end && i < b ){
			s++; (*ks)++;
			i++;
		}
		if( i < a )
			return FALSE;
		else
			return TRUE;

	case charsym:
		c = cc;
		next();
		match_repeater(&a, &b);
		i = 0;
		while( s < s_end && *s == c && i < b ){
			s++; (*ks)++;
			i++;
		}
		if( i < a )
			return FALSE;
		else
			return TRUE;
	
	default:
		__re_err("unexpected char %c", cc==0? ' ':cc);
	}
	return 0; /* per far contento gcc -Wall */
}


int match_continue(unsigned char *s, int *ks, int toplevel)
/*
	Esegue il parsing di una sequenza di zero o piu' branch:

		[ branch { ["|"] branch } ]
	
	toplevel vale 1 se siamo al livello di subex 0, cioe' in pratica se
	la funzione e' stata chiamata dalla funzione principale. Negli altri
	casi vale 0. Se toplevel==1 le subex potrebbero essere richieste
	dall'utilizzatore per recuperare sottostrighe da s; al contrario,
	se toplevel==0 allora siamo in una subex e possiamo aspettarci
	di trovare una ')' a chiusura della subex.

	Ritorna in *ks il n. di char si s consumati.
	Ritorna 1 se ha successo, 0 se non ha successo, -1 su errore di
	sintassi nella re.
*/
{
	int res, ks2;

	*ks = 0;
	do {
		if( m2_sym == endsym || (toplevel == 0 && m2_sym == rroundsym ) )
			return TRUE;
		res = match_branch(s, &ks2, toplevel);
		s += ks2; *ks += ks2;

		/* se res==1 il match con il branch corrente ha avuto successo:
		   => skip delle eventuali alternative "|" e continua;
		   se res==0 il match con il branch corrente non ha avuto successo:
		   => considera le eventuali alternative oppure fallisci: */
		if( res == 1 ){
			while( m2_sym == barsym ){
				next();
				match_skip_branch();
			}
		} else {
			if( m2_sym == barsym ){
				next();
			} else {
				return FALSE;
			}
		}
	}while( 1 );
	return TRUE;
}


int __match(unsigned char *_s_begin, unsigned char *_s_end, unsigned char *_re_begin, unsigned char* _re_end)
/*
	Esegue il matching tra la stringa s e la regular exp. re.
	Ritorna 1 se il match ha successo, 0 se non ha successo, -1 in
	caso di errore di sintassi nella re.

	ATTENZIONE! *_s_begin deve essere il primo carattere della stringa,
	mentre _s_end punta all'ultimo carattere della stringa +1;
	similmente per la re.
*/
{
	unsigned char *s, *re;
	int ks;

	re_begin = _re_begin;
	re = _re_begin;
	re_end = _re_end;
	s_begin = _s_begin;
	s = _s_begin;
	s_end = _s_end;
	parser_init(re);
	if( m2_sym == circsym ){
		next();
		return match_continue(s, &ks, 1);
	} else {
		/* devo trovare il primo match di *re in s: */
		while( s < s_end ){
			if( match_continue(s, &ks, 1) )  return 1;
			s++;
			parser_init(re);
		}
		return FALSE;
	}
	return 0; /* per far contento gcc -Wall */
}


int m2_match(STRING *s, STRING *re)
{
	int res;

	if( s == NULL || s->s == NULL )  return FALSE;
	if( re == NULL || re->s == NULL )  return FALSE;
	match_array = 0;
	m_a = NULL;
	res = __match(s->s, s->s + s->len, re->s, re->s + re->len);
	return res;
}


int m2_match_array(STRING *s, STRING *re, ARRAY **a)
{
	int res;

	if( s == NULL || s->s == NULL )  return FALSE;
	if( re == NULL || re->s == NULL )  return FALSE;
	match_array = 1;
	*a = NULL;
	m_a = a;
	res = __match(s->s, s->s + s->len, re->s, re->s + re->len);
	return (*a == NULL)? 0:(*m_a)->n_elems;
}

#endif
